from typing import Dict, Type

from django.contrib.postgres import fields as postgres_fields
from django.db.models import fields
from django.db.models.functions import Cast

from .expressions import ToIsoDate, ToIsoDateTime
from .settings import api_settings


class FieldMetaclass(type):
    def __new__(cls, name, bases, attrs):
        dj_fields = attrs.get("django_field_types", ())
        kls = super().__new__(cls, name, bases, attrs)
        kls._field_types_mapping.update({dj_field: kls for dj_field in dj_fields})
        return kls


# Abstract base field
class Field(metaclass=FieldMetaclass):
    _field_types_mapping = {}
    django_field_types = ()
    cast_type = None

    def __init__(self, source: str = None):
        self._source = source
        self.field_name = None
        self.serializer = None

    @property
    def source(self):
        return self._source or self.field_name

    def bind(self, field_name: str, serializer):
        self.field_name = field_name
        self.serializer = serializer

    def get_cast_type_params(self) -> Dict:
        return {}

    def expression(self):
        output_field = self.cast_type(**self.get_cast_type_params())
        return {self.field_name: Cast(self.source, output_field=output_field)}


# Base Json fields
class StringField(Field):
    django_field_types = (fields.CharField, fields.TextField, fields.UUIDField)
    cast_type = fields.TextField


class BooleanField(Field):
    django_field_types = (
        fields.BooleanField,
        fields.NullBooleanField,
    )
    cast_type = fields.BooleanField


class IntField(Field):
    django_field_types = (fields.IntegerField,)
    cast_type = fields.IntegerField


class FloatField(Field):
    django_field_types = (fields.FloatField,)
    cast_type = fields.FloatField


class DecimalField(Field):
    django_field_types = (fields.DecimalField,)
    cast_type = fields.DecimalField

    def __init__(self, coerce_to_string: bool = None, **kwargs):
        super().__init__(**kwargs)
        self.coerce_to_string = (
            coerce_to_string
            if coerce_to_string is not None
            else api_settings.COERCE_DECIMAL_TO_STRING
        )

    def get_cast_type_params(self) -> Dict:
        base_field = self.serializer._meta.model._meta.get_field(field_name=self.source)
        return {
            "max_digits": base_field.max_digits,
            "decimal_places": base_field.decimal_places,
        }

    def expression(self):
        if self.coerce_to_string:
            return {self.field_name: Cast(self.source, output_field=fields.TextField())}
        return super().expression()


class ObjectField(Field):
    django_field_types = (postgres_fields.JSONField,)
    cast_type = postgres_fields.JSONField


class ArrayField(Field):
    django_field_types = (postgres_fields.ArrayField,)
    cast_type = postgres_fields.ArrayField

    def get_cast_type_params(self) -> Dict:
        base_field = self.serializer._meta.model._meta.get_field(field_name=self.source)
        return {"base_field": base_field}


# Custom fields
class DateField(StringField):
    django_field_types = (fields.DateField,)
    cast_type = fields.TextField

    def expression(self):
        return {self.field_name: ToIsoDate(expression=self.source)}


class DateTimeField(StringField):
    django_field_types = (fields.DateTimeField,)
    cast_type = fields.TextField

    def expression(self):
        return {self.field_name: ToIsoDateTime(expression=self.source)}


def get_json_field_class_from_django_field(dj_field_class: Type[fields.Field]):
    # 1) try to get field class directly
    if json_field_class := Field._field_types_mapping.get(dj_field_class):
        return json_field_class

    # try to get a closest type
    for field_class, json_field in reversed(Field._field_types_mapping.items()):
        if issubclass(dj_field_class, field_class):
            return json_field

    raise ValueError(f"No JSON type is registered for {dj_field_class}.")



from django.conf import settings
from django.test.signals import setting_changed

from rest_framework.settings import APISettings as DrfAPISettings

DJANGO_SETTINGS_KEY = "PG_SERIALIZER"

DEFAULTS = {
    "COERCE_DECIMAL_TO_STRING": True,
    "JSON_AS_BYTES": False,
}


# List of settings that may be in string import notation.
IMPORT_STRINGS = ()


# List of settings that have been removed
REMOVED_SETTINGS = ()


class APISettings(DrfAPISettings):
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = self.__check_user_settings(user_settings)
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, DJANGO_SETTINGS_KEY, {})
        return self._user_settings

    def __check_user_settings(self, user_settings):
        for setting in REMOVED_SETTINGS:
            if setting in user_settings:
                raise RuntimeError(
                    f"The '{setting}' setting has been removed. "
                    "Please refer to the documentation for available settings."
                )
        return user_settings


api_settings = APISettings(None, DEFAULTS, IMPORT_STRINGS)


def reload_api_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == DJANGO_SETTINGS_KEY:
        api_settings.reload()


setting_changed.connect(reload_api_settings)



from django.db import models


class ToChar(models.Func):
    function = "TO_CHAR"
    output_field = models.TextField()
    text_format = None

    def __init__(self, expression: models.Expression, text_format: str = None, **extra):
        super().__init__(
            expression, models.Value(text_format or self.text_format), **extra,
        )


class ToIsoDate(ToChar):
    text_format = "YYYY-MM-DD"


class ToIsoDateTime(ToChar):
    text_format = 'YYYY-MM-DD"T"HH24:MI:SS.USTZH:TZM'



from contextlib import contextmanager
from functools import reduce
from typing import List, Dict, Tuple

import psycopg2.extensions
from django.db import connection, models


def dict_merge(*dicts: List[Dict]) -> Dict:
    return reduce(lambda acc, d: {**acc, **(d or {})}, dicts, {})


def generate_custom_select_with_params(
    queryset: models.QuerySet, annotations: Dict[str, models.Expression]
) -> Tuple[str, Tuple]:
    """Force Django to generate a custom select statement.
    Using internal query API:
    1) clear out all the default queryset fields, and remove added annotations
    2) recreate the internal annotations register using the provided ones
    3) generate sql + params to escape. params will be escaped by psycopg2 later.

    Parameters
    ----------
    queryset: models.QuerySet
    annotations: Dict[str, models.Expression]

    Returns
    -------
    Tuple[str, Tuple]

    """
    queryset.query.default_cols = False
    queryset.query.annotations = {}

    for alias, annotation in annotations.items():
        queryset.query.add_annotation(annotation, alias, is_summary=False)

    return queryset.query.sql_with_params()


@contextmanager
def db_str_as_bytes():
    try:
        psycopg2.extensions.register_type(
            psycopg2.extensions.BYTES, connection.connection
        )
        psycopg2.extensions.register_type(
            psycopg2.extensions.BYTESARRAY, connection.connection
        )
        yield
    finally:
        psycopg2.extensions.register_type(
            psycopg2.extensions.UNICODE, connection.connection
        )
        psycopg2.extensions.register_type(
            psycopg2.extensions.UNICODEARRAY, connection.connection
        )



from operator import attrgetter
from typing import Any, Dict, Sequence, Set, Tuple, Type

from django.db import connection, models, ProgrammingError
from django.utils.functional import cached_property

from .fields import Field, get_json_field_class_from_django_field
from .settings import api_settings
from .utils import dict_merge, generate_custom_select_with_params, db_str_as_bytes


class SerializerMetaclass(type):
    @classmethod
    def _get_declared_fields_from_bases(cls, bases: Sequence) -> Dict[str, Field]:
        def _has_base_fields(base):
            return (
                isinstance(base, cls)
                and hasattr(base, "_meta")
                and hasattr(base._meta, "declared_fields")
            )

        bases_fields = [
            base._meta.fields for base in reversed(bases) if _has_base_fields(base=base)
        ]
        return dict_merge(*bases_fields)

    @classmethod
    def _get_declared_fields(
        cls,
        name: str,
        bases: Sequence,
        attrs: Dict[str, Any],
        meta_fields: Sequence[str],
    ) -> Dict[str, Field]:
        bases_fields = cls._get_declared_fields_from_bases(bases=bases)
        local_fields = {
            name: attrs.pop(name)
            for name, obj in list(attrs.items())
            if isinstance(obj, Field)
        }

        if any(field not in meta_fields for field in local_fields):
            raise ProgrammingError(
                f"[{name}] One or more declared fields are not present in Meta.fields"
            )
        return {**bases_fields, **local_fields}

    @classmethod
    def _generate_auto_fields(
        cls, model: Type[models.Model], field_names: Set[str]
    ) -> Dict[str, Field]:
        return {
            field_name: get_json_field_class_from_django_field(
                dj_field_class=model._meta.get_field(field_name=field_name).__class__
            )()
            for field_name in field_names
        }

    def __new__(cls, name, bases, attrs):
        if name == "ModelSerializer":
            return super().__new__(cls, name, bases, attrs)

        model, meta_fields = attrgetter("model", "fields")(attrs["Meta"])

        declared_fields = cls._get_declared_fields(
            name=name, bases=bases, attrs=attrs, meta_fields=meta_fields
        )
        auto_fields = cls._generate_auto_fields(
            model=model,
            field_names=set(meta_fields).difference(set(declared_fields.keys())),
        )
        all_fields = {**declared_fields, **auto_fields}
        fields = {field: all_fields[field] for field in meta_fields}

        kls = super().__new__(cls, name, bases, attrs)

        # bind fields to this serializer
        for field_name, field_obj in fields.items():
            field_obj.bind(field_name=field_name, serializer=kls)

        # add back useful attrs
        meta_attrs = {
            "model": model,
            "declared_fields": declared_fields,
            "fields": fields,
        }
        setattr(kls, "_meta", type("Meta", (), meta_attrs))

        return kls


class ModelSerializer(metaclass=SerializerMetaclass):
    def __init__(self, queryset: models.QuerySet, as_bytes: bool = None):
        self.queryset = queryset
        self.as_bytes = as_bytes if as_bytes is not None else api_settings.JSON_AS_BYTES

    @property
    def query_with_params(self) -> Tuple[str, Dict]:
        field_expressions = dict_merge(
            *[field.expression() for field in self._meta.fields.values()]
        )
        sql, params = generate_custom_select_with_params(
            queryset=self.queryset, annotations=field_expressions
        )

        query = f"""
        WITH query AS ({sql})
        SELECT ARRAY_TO_JSON(ARRAY_AGG(ROW_TO_JSON(query)))::TEXT FROM query
        """
        return query, params

    def _run_query(self):
        with connection.cursor() as cursor:
            cursor.execute(*self.query_with_params)
            json_bytes = cursor.fetchone()[0]

        return json_bytes

    def to_representation(self) -> str or bytes:
        if self.as_bytes:
            with db_str_as_bytes():
                return self._run_query()
        else:
            return self._run_query()

    @cached_property
    def json(self) -> str or bytes:
        return self.to_representation()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pg_serializer']

package_data = \
{'': ['*']}

install_requires = \
['django', 'djangorestframework', 'psycopg2']

setup_kwargs = {
    'name': 'pg-serializer',
    'version': '0.1.1',
    'description': 'Fast serializers for Django using Postgres database',
    'long_description': '# pg_serializer: Very fast JSON serializer using Django and PostgreSQL\n\n[![PyPI version](https://badge.fury.io/py/pg-serializer.svg)](https://pypi.org/project/pg-serializer/)\n[![Python Version](https://img.shields.io/badge/python-3.8-blue.svg)](https://docs.python.org/3/whatsnew/3.8.html)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![pipeline status](https://gitlab.com/kozlek/pg_serializer/badges/master/pipeline.svg)](https://gitlab.com/kozlek/pg_serializer/-/commits/master)\n[![coverage report](https://gitlab.com/kozlek/pg_serializer/badges/master/coverage.svg)](https://gitlab.com/kozlek/pg_serializer/-/commits/master)\n\n## Overview \n\n> Django REST framework is a powerful and flexible toolkit for building Web APIs.\n\nDjango Rest Framework is a great framework for building REST APIs on top of Django and \nit fits 99% of use cases.  \nHowever, JSON serialization is done in Python and it can be very slow for \nsome `list` endpoints that return a lot of records.\n\n`pg_serializer` aims to soften this 1% issue by providing a way to do JSON serialization \ninside your PostgreSQL database. Indeed, PostgreSQL has a built-in support for JSON \nsince version 9.2, and it is far more faster than Python\'s one. \n\n## Installation\n\n```bash\npip install pg-serializer\n```\n\n## Basic usage\n\n`pg_serializer` features an automatic `ModelSerializer` like DRF.\n\n```python\nfrom django.contrib.auth.models import User\n\nimport pg_serializer\n\n\nclass UserSerializer(pg_serializer.ModelSerializer):\n    class Meta:\n        model = User\n        fields = (\n            "id",\n            "username",\n            "email",\n            "first_name",\n            "last_name",\n            "is_staff",\n            "date_joined",\n        )\n\n\njson_str = UserSerializer(User.objects.all()).json\n```\n\nIt also has preliminary support for ForeignKey relation using the Django `__` separator.\n\n```python\nimport pg_serializer\n\n\nclass OrderSerializer(pg_serializer.ModelSerializer):\n    buyer = pg_serializer.StringField(source="buyer__username")\n\n    class Meta:\n        model = Order\n        fields = (\n            "transaction_id",\n            "transaction_time",\n            "buyer",\n            "is_gift",\n            "shipping_date",\n            "additional_data",\n        )\n\n\njson_str = OrderSerializer(Order.objects.all()).json\n```\n\nFull examples are available inside `tests/models.py` and `tests/serializers.py`.\n\n\n## Disclaimer ⚠️\n\n- `pg_serializer` is not designed to replace DRF serializers, only to speed up \nsome endpoints when performance is becoming an issue\n- `pg_serializer` is still in alpha development: bugs can occur and \nit doesn\'t support all Django fields and relations\n\n## Roadmap\n\n- reinforce the test suite\n- implement (and document) custom datetime and decimal formatting\n- document how to create custom serializer fields\n- full support for ArrayField and JSONField\n- complex queryset operations (GROUP BY, ...) support\n',
    'author': 'Thomas Berdy',
    'author_email': 'thomas.berdy@outlook.com',
    'maintainer': 'Thomas Berdy',
    'maintainer_email': 'thomas.berdy@outlook.com',
    'url': 'https://gitlab.com/kozlek/pg_serializer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
